/*
 * Windows Registry FUSE Filesystem
 *
 * Version number definitions
 *
 */

#ifndef WINREGFS_VERSION_H
#define WINREGFS_VERSION_H
#define VER "0.7"
#define VERDATE "2017-03-10"
#endif	/* WINREGFS_VERSION_H */
