#ifndef CONFIG_H
#define CONFIG_H

/* Threaded mode suffers from decreased performance */
#define ENABLE_THREADED 0

#define ENABLE_LOGGING 0
#define ENABLE_DEBUG_LOGGING 0
#define ENABLE_DEBUG_PRINTF 0

#define ENABLE_NKOFS_CACHE 1
#define ENABLE_NKOFS_CACHE_STATS 0
#define NKOFS_CACHE_ITEMS 64

#endif	/* CONFIG_H */
